/*
 * DO NOT ALTER OR REMOVE COPYRIGHT NOTICES OR THIS HEADER.
 * 
 * Copyright 2011 OpenConcerto, by ILM Informatique. All rights reserved.
 * 
 * The contents of this file are subject to the terms of the GNU General Public License Version 3
 * only ("GPL"). You may not use this file except in compliance with the License. You can obtain a
 * copy of the License at http://www.gnu.org/licenses/gpl-3.0.html See the License for the specific
 * language governing permissions and limitations under the License.
 * 
 * When distributing the software, include this License Header Notice in each file.
 */
 
 package org.openconcerto.erp.core.sales.pos.ui;

import org.openconcerto.erp.config.ComptaPropsConfiguration;
import org.openconcerto.erp.config.Gestion;
import org.openconcerto.erp.core.sales.pos.POSConfiguration;
import org.openconcerto.erp.core.sales.pos.model.Article;
import org.openconcerto.erp.core.sales.pos.model.Client;
import org.openconcerto.erp.core.sales.pos.model.DBState;
import org.openconcerto.erp.core.sales.pos.model.ReceiptCode;
import org.openconcerto.erp.core.sales.pos.model.RegisterDB;
import org.openconcerto.erp.core.sales.pos.model.RegisterFiles;
import org.openconcerto.erp.core.sales.pos.model.RegisterLog;
import org.openconcerto.erp.core.sales.pos.model.RegisterState;
import org.openconcerto.erp.core.sales.pos.model.RegisterState.Status;
import org.openconcerto.erp.core.sales.pos.model.Ticket;
import org.openconcerto.sql.PropsConfiguration;
import org.openconcerto.sql.RemoteShell;
import org.openconcerto.sql.element.SQLElementDirectory;
import org.openconcerto.sql.model.SQLBase;
import org.openconcerto.sql.model.SQLRowAccessor;
import org.openconcerto.sql.model.SQLRowValues;
import org.openconcerto.sql.sqlobject.ElementComboBox;
import org.openconcerto.utils.ClassPathLoader;
import org.openconcerto.utils.ExceptionHandler;
import org.openconcerto.utils.cc.ExnTransformer;

import java.awt.Container;
import java.awt.Dimension;
import java.awt.Frame;
import java.awt.Point;
import java.awt.Toolkit;
import java.awt.event.MouseEvent;
import java.io.File;
import java.io.IOException;
import java.net.MalformedURLException;
import java.sql.SQLException;
import java.text.ParseException;
import java.util.EnumSet;
import java.util.List;
import java.util.Set;
import java.util.concurrent.Callable;
import java.util.concurrent.FutureTask;
import java.util.logging.Level;

import javax.swing.JFrame;
import javax.swing.JOptionPane;
import javax.swing.JPanel;
import javax.swing.SwingUtilities;
import javax.swing.ToolTipManager;

import org.jdom2.JDOMException;

public class CaisseFrame extends JFrame {
    private final ComptaPropsConfiguration conf;
    private final RegisterFiles files;
    private final RegisterDB registerDB;
    final CaissePanel mainPanel;

    CaisseFrame(final ComptaPropsConfiguration conf, final RegisterFiles files, final RegisterDB registerDB) throws Exception {
        this.conf = conf;
        this.files = files;
        this.registerDB = registerDB;
        this.mainPanel = new CaissePanel(this);
        setContentPane(mainPanel);
        setFocusable(true);
    }

    public final ComptaPropsConfiguration getConf() {
        return this.conf;
    }

    public final RegisterFiles getFiles() {
        return this.files;
    }

    public final RegisterDB getDB() {
        return this.registerDB;
    }

    public static void main(String[] args) {
        System.setProperty(SQLRowAccessor.ACCESS_DB_IF_NEEDED_PROP, "true");
        try {
            System.out.println("Lancement du module de caisse");
            ToolTipManager.sharedInstance().setInitialDelay(0);
            RemoteShell.startDefaultInstance(null, null);

            System.setProperty(PropsConfiguration.REDIRECT_TO_FILE, "true");
            System.setProperty(SQLBase.ALLOW_OBJECT_REMOVAL, "true");

            ExceptionHandler.setForceUI(true);
            ExceptionHandler.setForumURL("http://www.openconcerto.org/forum");
            ExceptionHandler.setThrowableHandler(Gestion.createDefaultThrowableHandler());
            // SpeedUp Linux
            System.setProperty("sun.java2d.pmoffscreen", "false");
            System.setProperty(SQLBase.STRUCTURE_USE_XML, "true");
            System.setProperty(PropsConfiguration.REDIRECT_TO_FILE, "true");
            if (POSConfiguration.getInstance().isUsingJPos()) {
                ClassPathLoader c = ClassPathLoader.getInstance();
                try {
                    final List<String> posDirectories = POSConfiguration.getInstance().getJPosDirectories();
                    for (String posDirectory : posDirectories) {
                        if (posDirectory != null && !posDirectory.trim().isEmpty()) {
                            c.addJarFromDirectory(new File(posDirectory.trim()));
                        }
                    }
                } catch (MalformedURLException e) {
                    e.printStackTrace();
                }
                c.load();

            }
            
            // TODO Ajouter la gestion des arguments en ligne de commande
            
            final ComptaPropsConfiguration conf = POSConfiguration.getInstance().createConnexion();

            final int userID = POSConfiguration.getInstance().getUserID();
            final RegisterFiles registerFiles = new RegisterFiles(POSConfiguration.getInstance().getRootDir(), true, POSConfiguration.getInstance().getPosID());
            final RegisterDB registerDB = new RegisterDB(conf.getDirectory(), conf.getProductInfo(), POSConfiguration.getInstance().getPosID());

            // check before changing any state
            final boolean quit = registerDB.fetchRegisterState().checkIfMoved();
            if (quit) {
                quit();
                return;
            }

            final RegisterState reconciledState = registerFiles.doWithLock(new ExnTransformer<RegisterFiles, RegisterState, IOException>() {
                @Override
                public RegisterState transformChecked(RegisterFiles input) throws IOException {
                    try {
                        return reconcileFSandDB(conf.getDirectory(), input, registerDB, userID);
                    } catch (Exception e) {
                        throw new IOException("Couldn't reconcile local and remote state", e);
                    }
                }
            });
            POSConfiguration.getLogger().log(Level.INFO, "FS and DB states reconciled : {0}", reconciledState);

            System.setProperty("awt.useSystemAAFontSettings", "on");
            System.setProperty("swing.aatext", "true");
            System.setProperty(ElementComboBox.CAN_MODIFY, "true");

            RegisterState state = reconciledState;
            if (reconciledState.getStatus() == Status.CLOSED) {
                final FutureTask<Boolean> askUserCallable = new FutureTask<>(new Callable<Boolean>() {
                    @Override
                    public Boolean call() throws Exception {
                        final int ans = JOptionPane.showConfirmDialog(null, "La caisse n’est pas ouverte, voulez-vous l’ouvrir ?", "Caisse fermée", JOptionPane.YES_NO_OPTION);
                        return ans == JOptionPane.YES_OPTION;

                    }
                });
                SwingUtilities.invokeLater(askUserCallable);
                final boolean userAgreed = askUserCallable.get();
                if (userAgreed) {
                    final RegisterLog newLog = registerFiles.open(userID, registerDB);
                    state = newLog.getRegisterState();
                }
            }
            if (state.getStatus() != Status.OPEN) {
                POSConfiguration.getLogger().log(Level.FINE, "State not open ({0}), exiting", state);
                POSConfiguration.getInstance().closeConnexion();
                return;
            }
            POSConfiguration.getLogger().log(Level.INFO, "FS and DB states open, opening UI");

            SwingUtilities.invokeLater(new Runnable() {
                public void run() {

                    try {
                        // UIManager.setLookAndFeel(UIManager.getSystemLookAndFeelClassName());
                    } catch (Exception e) {
                        e.printStackTrace();
                    }
                    try {
                        CaisseFrame f = new CaisseFrame(conf, registerFiles, registerDB);
                        f.setUndecorated(true);
                        f.setDefaultCloseOperation(JFrame.EXIT_ON_CLOSE);

                        f.pack();
                        if (System.getProperty("os.name").toLowerCase().startsWith("mac os x")) {
                        	f.setLocation(0, 24);
                        } else {
                        	f.setLocation(0, 0);
                        }
                        Dimension screenSize = Toolkit.getDefaultToolkit().getScreenSize();
                        if (POSConfiguration.getInstance().getScreenWidth() > 0 && POSConfiguration.getInstance().getScreenHeight() > 0) {
                            f.setSize(new Dimension(POSConfiguration.getInstance().getScreenWidth() - f.getX(), 
                            		POSConfiguration.getInstance().getScreenHeight() - f.getY()));
                        } else {
                            f.setSize(new Dimension(screenSize.getSize().width - f.getX() , screenSize.getSize().height - f.getY()));//screenSize);
                        }
                        System.out.println("Affichage de l'interface");
                        f.setVisible(true);
                        if (screenSize.getWidth() < 1280 || screenSize.getHeight() < 720) {
                            JOptionPane.showMessageDialog(f,
                                    "La résolution de votre écran est trop faible.\nLa largeur doit être au minium de 1280 pixels.\nLa hauteur doit être au minium de 720 pixels.");
                        }
                    } catch (Throwable e) {
                        // Catch throwable to be able to see NoClassDefFound and other hard issues
                        ExceptionHandler.handle("Erreur d'initialisation de la caisse (main)", e);
                    }

                }
            });
        } catch (Throwable e) {
            // Catch throwable to be able to see NoClassDefFound and other hard issues
            ExceptionHandler.handle("Erreur d'initialisation de la caisse", e);
        }
    }

    private static RegisterState reconcileFSandDB(final SQLElementDirectory dir, final RegisterFiles files, final RegisterDB registerDB, final int userID)
            throws IOException, JDOMException, ParseException, SQLException {

        // *** find local and remote states
        final RegisterLog lastLog = files.getLastLog();
        POSConfiguration.getLogger().log(Level.CONFIG, "Found last log {0}", lastLog);

        RegisterState localState;
        if (lastLog == null) {
            final List<File> remainingReceipts = ReceiptCode.getReceiptsToImport(files.getPosID());
            localState = new RegisterState(remainingReceipts.isEmpty() ? Status.CLOSED : Status.OBSOLETE, null);
        } else {
            localState = lastLog.getRegisterState();
        }
        POSConfiguration.getLogger().log(Level.CONFIG, "FS state {0}", localState);

        final DBState fetchedState = registerDB.fetchRegisterState();
        RegisterState remoteState = fetchedState.getRegisterState();
        POSConfiguration.getLogger().log(Level.CONFIG, "DB state {0}", fetchedState);

        // *** import remaining obsolete receipts
        if (localState.getStatus() == Status.OBSOLETE) {
            if (remoteState.hasDate())
                throw new IllegalStateException("There remains obsolete receipts but DB can no longer import them : " + remoteState);
            final List<Ticket> allTickets = POSConfiguration.getInstance().allTickets();
            POSConfiguration.getLogger().log(Level.INFO, "{0} obsolete receipt(s) will be stored in the DB", allTickets.size());
            POSConfiguration.getInstance().commitAll(allTickets);
            final List<File> remainingReceipts = ReceiptCode.getReceiptsToImport(files.getPosID());
            if (!remainingReceipts.isEmpty())
                throw new IllegalStateException("Not all obsolete receipts could be imported : " + remainingReceipts);
            localState = new RegisterState(Status.CLOSED, null);
            POSConfiguration.getLogger().log(Level.FINE, "All obsolete receipts have been stored in the DB");
        }
        final EnumSet<Status> validStatus = EnumSet.of(Status.OPEN, Status.CLOSED);
        if (!validStatus.contains(localState.getStatus()))
            throw new IllegalStateException("Unexpected local status :" + localState);
        if (!validStatus.contains(remoteState.getStatus()))
            throw new IllegalStateException("Unexpected remote status :" + remoteState);

        // *** reconcile if possible
        if (!localState.equals(remoteState)) {
            POSConfiguration.getLogger().log(Level.INFO, "Different FS and DB state, will try to reconcile\nFS " + localState + " with\nDB " + remoteState);

            // OK because of the check above
            final boolean localOpen = localState.getStatus() == Status.OPEN;
            final boolean remoteOpen = remoteState.getStatus() == Status.OPEN;
            // Time line
            // 0
            // Open DB
            // 1
            // Open Local
            // 2
            // Create tickets...
            // 3
            // Close Local
            // 4
            // Close DB
            // 5
            // Open DB
            // 6
            // Open Local
            // 7
            try {
                if (localOpen) {
                    if (remoteOpen) {
                        throw new IllegalStateException("Both open with but with different dates");
                    } else {
                        // DB is at 0 or 5, local is at 2
                        throw new IllegalStateException("local is open but the DB isn't");
                    }
                } else if (remoteOpen) {
                    assert remoteState.hasDate() : "Remote state open without date : " + remoteState;
                    final SQLRowValues lastClosure = fetchedState.getLastClosureEntry();
                    final java.util.Date lastClosureDate = lastClosure == null ? null : lastClosure.getDate("DATE").getTime();
                    if (!localState.hasDate()) {
                        // at 1
                        // MAYBE allow it for new install, for now the receipts must be copied
                        // over
                        if (lastClosureDate != null)
                            throw new IllegalStateException("DB was closed (" + lastClosureDate + ") and now open, but local log is missing");
                        try {
                            localState = files.open(userID, fetchedState).getRegisterState();
                        } catch (Exception e) {
                            throw new IllegalStateException("The local opening (following the already open DB) failed", e);
                        }
                    } else if (remoteState.compareDateTo(lastLog.getFirstRegisterEvent().getDate()) == 0) {
                        // at 4
                        try {
                            remoteState = registerDB.close(userID, files.getLastLog()).getRegisterState();
                        } catch (Exception e) {
                            throw new IllegalStateException("The closure of the DB (following the already closed local) failed", e);
                        }
                    } else if (lastClosureDate != null && localState.compareDateTo(lastClosureDate) == 0) {
                        // at 6, TODO factor with above
                        try {
                            localState = files.open(userID, fetchedState).getRegisterState();
                        } catch (Exception e) {
                            throw new IllegalStateException("The local opening (following the already open DB) failed", e);
                        }
                    } else {
                        // DB is at 1, local is between 4 and 6
                        throw new IllegalStateException("DB was opened for a different date");
                    }
                } else {
                    if (!localState.hasDate()) {
                        assert remoteState.hasDate() : "Both closed with no dates, but not equal";
                        // DB is at 5, local is at 0
                        // MAYBE allow it for new install, for now the receipts must be copied
                        // over
                        throw new IllegalStateException("DB was opened and closed, but local log is missing");
                    } else if (remoteState.hasDate()) {
                        // DB is at 5, local is at 5 but not for the same day
                        assert remoteState.compareDateTo(localState) != 0 : "Both closed with equal dates, but not equal";
                        throw new IllegalStateException("DB was opened and closed for a different date");
                    } else {
                        // DB is at 0, local is between 4 and 6
                        throw new IllegalStateException("DB was never opened but local was closed");
                    }
                }
            } catch (Exception e) {
                throw new IllegalStateException("States couldn't be reconciled, local : " + localState + ", remote : " + remoteState, e);
            }
        }
        if (!remoteState.equals(localState))
            throw new IllegalStateException("Unexpected state");
        return remoteState;
    }

    static public void quit() {
        POSConfiguration.getLogger().log(Level.INFO, "User exit");
        POSConfiguration.getInstance().closeConnexion();
        Frame[] l = Frame.getFrames();
        for (int i = 0; i < l.length; i++) {
            Frame f = l[i];
            System.err.println(f.getName() + " " + f + " Displayable: " + f.isDisplayable() + " Valid: " + f.isValid() + " Active: " + f.isActive());
        }
        Set<Thread> threadSet = Thread.getAllStackTraces().keySet();
        for (Thread thread : threadSet) {
            if (!thread.isDaemon()) {
                System.err.println(thread.getName() + " " + thread.getId() + " not daemon");
            }
        }
    }

    @Override
    public void dispose() {
        quit();
        // Fermeture
        this.getControler().setLCD("   CAISSE FERMEE    ", "", 0);
        super.dispose();
    }

    public void showMenu() {
        System.out.println("CaisseFrame.showMenu()");
        final CaisseMenuPanel panel = new CaisseMenuPanel(this);
        showPanel(panel);
        this.getControler().setLCD("OpenConcerto", "Menu", 0);

    }

    public void showPanel(JPanel panel) {
        this.invalidate();
        final int x = (getWidth() - panel.getPreferredSize().width) / 2;
        final int y = 100;
        final POSGlassPane glassPane2 = new POSGlassPane(panel, x, y) {
            @Override
            public void mousePressed(MouseEvent e) {
                Point containerPoint = SwingUtilities.convertPoint(this, e.getPoint(), panel);
                if (containerPoint.x < 0 || containerPoint.x > panel.getWidth() || containerPoint.y < 0 || containerPoint.y > panel.getHeight()) {
                    setGlassPane(new JPanel());
                    getGlassPane().setVisible(false);
                }
                super.mousePressed(e);
            }

        };
        this.setGlassPane(glassPane2);
        this.getGlassPane().setVisible(true);
        this.validate();
        this.repaint();
    }

    public void showPriceEditor(Article article, CaisseControler caisseControler) {
        getControler().disableBarcodeReader();
        System.out.println("CaisseFrame.showPriceEditor()");
        this.invalidate();
        final PriceEditorPanel panel = new PriceEditorPanel(this, article);

        final POSGlassPane glassPane2 = new POSGlassPane(panel, (getWidth() - panel.getPreferredSize().width) / 2, 100) {
            @Override
            public void mousePressed(MouseEvent e) {
                Point containerPoint = SwingUtilities.convertPoint(this, e.getPoint(), panel);
                if (containerPoint.x < 0 || containerPoint.x > panel.getWidth() || containerPoint.y < 0 || containerPoint.y > panel.getHeight()) {
                    setGlassPane(new JPanel());
                    getGlassPane().setVisible(false);
                    getControler().enableBarcodeReader();

                }
                super.mousePressed(e);
            }

        };
        this.setGlassPane(glassPane2);
        this.getGlassPane().setVisible(true);
        this.validate();
        this.repaint();

    }

    public void showCaisse() {
        getControler().enableBarcodeReader();
        setGlassPane(new JPanel());
        getGlassPane().setVisible(false);

        System.out.println("CaisseFrame.showCaisse()");
        this.setContentPane(this.mainPanel);
        this.getControler().setLCD("OpenConcerto", "Caisse", 0);
        this.getControler().setLCDDefaultDisplay(5);
    }

    public void showTickets(Ticket t) {
        System.out.println("CaisseFrame.showMenu()");
        final ListeDesTicketsPanel panel = new ListeDesTicketsPanel(this);
        panel.setSelectedTicket(t);
        this.setContentPane(panel);
    }

    public CaisseControler getControler() {
        return this.mainPanel.getControler();
    }

    public void showClients() {
        System.out.println("CaisseFrame.showClients()");
        final ListeDesClientsPanel panel = new ListeDesClientsPanel(this);
        this.setContentPane(panel);
    }

    public void setClient(Client client) {
        System.err.println("CaisseFrame.setClient() " + client.getFullName());
        this.getControler().setClient(client);

    }

    @Override
    public void setContentPane(Container contentPane) {
        this.invalidate();
        setGlassPane(new JPanel());
        getGlassPane().setVisible(false);
        super.setContentPane(contentPane);
        this.validate();
        this.repaint();
    }
}
